# MIT License
#
# Copyright (c) 2018 Evgeny Medvedev, evge.medvedev@gmail.com
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from ethereumetl.domain.contract import EthContract


class EthContractMapper(object):

    def rpc_result_to_contract(self, contract_address, rpc_result):
        contract = EthContract()
        contract.address = contract_address
        contract.bytecode = rpc_result

        return contract

    def contract_to_dict(self, contract):
        return {
            'type': 'contract',
            'address': contract.address,
            'bytecode': contract.bytecode,
            'function_sighashes': contract.function_sighashes,
            'is_erc20': contract.is_erc20,
            'is_erc721': contract.is_erc721,
            'block_number': contract.block_number
        }



#%%
from google.cloud import bigquery
import json

# %% For local
import google.auth
import os
HOME = os.getcwd()
CLIENT_ID = 'contxtsio-nftbank'
CREDENTIALS = f"./.cred/pipeline-executor.contxtsio-nftbank.credentials"
os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = CREDENTIALS

SCOPE = ('https://www.googleapis.com/auth/bigquery',
         'https://www.googleapis.com/auth/cloud-platform',
         'https://www.googleapis.com/auth/drive')
credentials, _ = google.auth.default(scopes=SCOPE)

bq_client = bigquery.Client(CLIENT_ID, credentials=credentials)

# %%
def load_table(data_name):
    schema_dicts = json.load(open(f"./schema/raw/{data_name}.json", "r+"))
    schema = [bigquery.SchemaField.from_api_repr(schema_dict) for schema_dict in schema_dicts]

    # Load bq data
    job_config = bigquery.LoadJobConfig()
    hive_partitioning = bigquery.external_config.HivePartitioningOptions()
    hive_partitioning.mode = "CUSTOM"
    hive_partitioning.source_uri_prefix = f"gs://nftbank-matic-network/export/raw/{data_name}/{{block_date:DATE}}/{{block_hour:STRING}}/"  #path to hive partition data
    job_config.hive_partitioning = hive_partitioning

    job_config.source_format = bigquery.SourceFormat.NEWLINE_DELIMITED_JSON
    job_config.create_disposition = bigquery.CreateDisposition.CREATE_IF_NEEDED
    job_config.write_disposition = bigquery.WriteDisposition.WRITE_APPEND
    job_config.schema = schema
    # job_config.time_partitioning = bigquery.TimePartitioning(type_=bigquery.TimePartitioningType().DAY, field='block_date')
    job_config.autodetect = True

    # # full
    # uri = "gs://nftbank-metadata/discord/users/*.jsonl"  #path to data

    # incr
    uri = f"gs://nftbank-matic-network/export/raw/{data_name}/*.json"  #path to data

    load_job = bq_client.load_table_from_uri(uri,
                                        f"contxtsio-nftbank.matic_staging_area.raw_{data_name}",
                                        job_config=job_config)  # API request

    load_job.result()  # Waits for table load to complete.


# %%
load_table('blocks')
load_table('transactions')
load_table('logs')
load_table('receipts')



import logging
import threading
import time


class RateLimitingProxy:
    def __init__(self, delegate, max_per_second=5):
        self._lock = threading.Lock()
        self._delegate = delegate
        self._min_interval = 1.0 / max_per_second
        self._last_time_called = time.perf_counter()
        self._wait_buffer = 0.001

    def __getattr__(self, name):
        self._lock.acquire()
        try:
            elapsed = time.perf_counter() - self._last_time_called
            left_to_wait = self._min_interval - elapsed + self._wait_buffer
            if left_to_wait > 0:
                logging.debug(f'Waiting {left_to_wait} seconds because of rate limiting')
                time.sleep(left_to_wait)

            return getattr(self._delegate, name)
        finally:
            self._last_time_called = time.perf_counter()
            self._lock.release()


#%%
from datetime import datetime, timedelta, date
from typing import List, Union, Iterator

#%%
def daterange(start_date: datetime,
              end_date: datetime,
              interval: timedelta = timedelta(days=1)) -> Iterator[datetime]:

    for i in range(0, int((end_date - start_date).total_seconds()), int(interval.total_seconds())):
        yield start_date + timedelta(seconds=i)


def chunks(it, n):
    chunk = []
    for i in it:
        chunk.append(i)
        if len(chunk) >= n:
            yield chunk
            chunk = []

    if len(chunk) > 0:
        yield chunk

#%%
def timestamp_chunk(start_timestamp: datetime,
                    end_timestamp: datetime,
                    interval: timedelta = timedelta(hours=1)) -> Iterator[datetime]:

    for i in range(int(interval.total_seconds()), int((end_timestamp - start_timestamp).total_seconds()), int(interval.total_seconds())):
        yield start_timestamp + timedelta(seconds=i-3600), start_timestamp + timedelta(seconds=i)

# %%



# %%
import pymongo
from pymongo import ReplaceOne
from bson.codec_options import CodecOptions

from datetime import datetime, timezone, timedelta
from typing import List, Any, Union, Iterator, Dict
from copy import deepcopy
from utils.commons import timestamp_chunk

# %%
TaskIDs = {
    "BLOCKS_AND_TRANSACTIONS": "matic_raw_blocks_and_transactions",
    "RECEIPTS_AND_LOGS": "matic_raw_receipts_and_logs",
    "CONTRACTS": "matic_raw_contracts",
    "TOKENS": "matic_raw_tokens",
    "TOKEN_TRANSFERS": "matic_raw_token_transfers",
    "TRACES": "matic_raw_traces",
    "ALL": "matic_all"
}

# %%
class TaskManager:
    ALLOWED_STAGES = ["raw", "enrich"]

    _connection_uri:str = None
    _client:pymongo.MongoClient = None
    _db:pymongo.database.Database = None
    _min_timestamp:datetime = None
    _max_timestamp:datetime = None

    def __init__(self, mongodb_connection_uri:str, min_timestamp:datetime, max_timestamp:datetime=None):

        if mongodb_connection_uri is not None:
            self._connection_uri = mongodb_connection_uri
            self._client = pymongo.MongoClient(
                self._connection_uri
            )
            self._db = self._client["matic_etl"]

        self._min_timestamp = min_timestamp
        if max_timestamp is not None:
            self._max_timestamp = max_timestamp
        else:
            self._max_timestamp = datetime.now(tz=timezone.utc)

    def get_client(self):
        return self._client

    def get_db(self):
        return self._db

    def initialize_db(self, task_id, stage):
        options = CodecOptions(tz_aware=True)
        get_last_tasks = list(self._db.get_collection(f"tasks_{stage}", codec_options=options).find({
            "task_id": task_id,
        }, {
            "_id": 0,
            "end_timestamp": 1
        }).sort([('start_timestamp', -1)]).limit(1))

        if (len(get_last_tasks) > 0):
            _min_timestamp = get_last_tasks[0]["end_timestamp"]
        else:
            _min_timestamp = self._min_timestamp

        job_ranges = timestamp_chunk(
            start_timestamp=max(self._min_timestamp, _min_timestamp),
            end_timestamp=self._max_timestamp,
            interval=timedelta(hours=1))

        tasks = ({
            "_id": f"""{task_id}_{int(start_timestamp.timestamp())}_{int(end_timestamp.timestamp())}""",
            "task_id": task_id,
            "start_timestamp": start_timestamp,
            "end_timestamp": end_timestamp,
        } for start_timestamp, end_timestamp in job_ranges)

        upserts=[ReplaceOne({'_id':x['_id']}, x, upsert=True) for x in tasks]

        if len(upserts) > 0:
            options = CodecOptions(tz_aware=True)
            self._db.get_collection(f"tasks_{stage}", codec_options=options).bulk_write(upserts)


    def get_todos(self, task_id:str, stage:str, limit:int=48):
        options = CodecOptions(tz_aware=True)
        return list(self._db.get_collection(f"tasks_{stage}", codec_options=options).find({
            "task_id": task_id,
            "finished_at": { "$exists": False }
        }, {
            "_id": 0,
            "start_timestamp": 1,
            "end_timestamp": 1
        }).sort([('start_timestamp', 1)]).limit(limit))


    def task_finished(self,
                      task_id:str,
                      stage:str,
                      start_timestamp:datetime,
                      end_timestamp:datetime):
        options = CodecOptions(tz_aware=True)
        self._db.get_collection(f"tasks_{stage}", codec_options=options).find_one_and_update({
                "_id": f"""{task_id}_{int(start_timestamp.timestamp())}_{int(end_timestamp.timestamp())}"""
            }, {
                "$set": {
                    "finished_at": datetime.now(tz=timezone.utc)
                }
            })

    def task_started(self,
                      task_id:str,
                      stage:str,
                      start_timestamp:datetime,
                      end_timestamp:datetime):

        options = CodecOptions(tz_aware=True)
        self._db.get_collection(f"tasks_{stage}", codec_options=options).find_one_and_replace({
                "_id": f"""{task_id}_{int(start_timestamp.timestamp())}_{int(end_timestamp.timestamp())}""",
            }, {
                "_id": f"""{task_id}_{int(start_timestamp.timestamp())}_{int(end_timestamp.timestamp())}""",
                "task_id": task_id,
                "start_timestamp": start_timestamp,
                "end_timestamp": end_timestamp,
                "started_at": datetime.now(tz=timezone.utc)
            }, upsert=True)

    def get_latest_job_result(self, task_id:str):
        # TODO: get from MongoDB
        return {
            "task_id": task_id,
            "start_timestamp": datetime(2021, 3, 1, 23, 0, 0, tzinfo=timezone.utc),
            "end_timestamp": datetime(2021, 3, 2, 0, 0, 0, tzinfo=timezone.utc),
            "started_at": datetime(2021, 3, 2, 0, 3, 0, tzinfo=timezone.utc),
            "finished_at": datetime(2021, 3, 2, 0, 5, 32, tzinfo=timezone.utc)
        }


    def set_latest_job_result(
            self,
            task_id:str,
            start_timestamp:datetime,
            end_timestamp:datetime,
            started_at:Union[None, datetime]=None,
            finished_at:Union[None, datetime]=None):

        # TODO: set to MongoDB
        if started_at is not None and isinstance(started_at, datetime):
            print({
                "_id": f"""{task_id}_{int(start_timestamp)}_{int(end_timestamp)}""",
                "task_id": task_id,
                "start_timestamp": start_timestamp,
                "end_timestamp": end_timestamp,
                "started_at": datetime.now(tz=timezone.utc)
            })
        elif started_at is not None and isinstance(started_at, datetime):
            print({
                "_id": f"""{task_id}_{int(start_timestamp)}_{int(end_timestamp)}""",
                "task_id": task_id,
                "start_timestamp": start_timestamp,
                "end_timestamp": end_timestamp,
                "finished_at": finished_at.replace(tz=timezone.utc)
            })
        else:
            print({})



# %%
from google.cloud import storage
import google.auth
import os

# %%
class SimpleGoogleCloudStorageHook:

    _conn = None
    _project_id = None
    _credentials = None
    MEGABYTE = 1024 * 1024

    def __init__(self, credentials, project_id):
        self._credentials = credentials
        self._project_id = project_id
        self._conn = storage.Client(credentials=self._credentials,
                                    project=self._project_id)

    def get_conn(self):
        if self._conn is None:
            self._conn = storage.Client(credentials=self._credentials,
                                        project=self._project_id)
        return self._conn

    # Helps avoid OverflowError: https://stackoverflow.com/questions/47610283/cant-upload-2gb-to-google-cloud-storage
    # https://developers.google.com/api-client-library/python/guide/media_upload#resumable-media-chunked-upload
    def upload_to_gcs(self, gcs_hook, bucket, object, filename, mime_type='application/octet-stream'):
        storage_client = gcs_hook.get_conn()

        bucket = storage_client.bucket(bucket)

        if os.path.getsize(filename) > 10 * self.MEGABYTE:
            blob = bucket.blob(object, chunk_size=10 * self.MEGABYTE)
        else:
            blob = bucket.blob(object)

        blob.upload_from_filename(filename)


    # Can download big files unlike gcs_hook.download which saves files in memory first
    def download_from_gcs(self, gcs_hook, bucket, object, filename):
        storage_client = gcs_hook.get_conn()

        bucket = storage_client.get_bucket(bucket)
        blob_meta = bucket.get_blob(object)

        if blob_meta.size > 10 * self.MEGABYTE:
            blob = bucket.blob(object, chunk_size=10 * self.MEGABYTE)
        else:
            blob = bucket.blob(object)

        blob.download_to_filename(filename)



# %%
# Append parent package (DEV)
import os, sys
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), '..'))


# %%
import logging


# %%
from ethereumetl.cli import (
    get_block_range_for_timestamps,
    extract_field,
    export_blocks_and_transactions,
    export_receipts_and_logs,
    extract_contracts,
    extract_tokens,
    extract_token_transfers,
    export_geth_traces,
    extract_geth_traces,
)
import os


# %%
from datetime import datetime, timedelta, timezone
from tempfile import TemporaryDirectory
from typing import List, Union, Iterator


# %%
from utils.commons import timestamp_chunk
from utils.google_cloud_storage_hook import SimpleGoogleCloudStorageHook
from utils.task_manager import TaskManager, TaskIDs

# %%
import httplib2
import google.auth
import google_auth_httplib2
import google.oauth2.service_account
from google.api_core.exceptions import GoogleAPICallError, AlreadyExists, RetryError
from googleapiclient.errors import HttpError


# %%
provider_uris = [
    "https://rpc-mainnet.maticvigil.com/v1/927c67086ea722f1b98b3ec14b174ce140d3a84e",
    "https://rpc-mainnet.maticvigil.com/v1/5dc314d2597052c4d1e00370733d762e5fcd3334",
    "https://rpc-mainnet.maticvigil.com/v1/908b40606eada9119d934ae2809b566b87f61cd7",
    "https://rpc-mainnet.maticvigil.com/v1/058a788a199c0017ce46b3900ddd801c6412372b",
    "https://rpc-mainnet.maticvigil.com/v1/3c91e94ed992e48ae091ed12d1d821f2fbdc0b43",
    "https://rpc-mainnet.maticvigil.com/v1/f9c8eaeeb3c203c8fab2a978c41300b68421dccc",
    "https://rpc-mainnet.maticvigil.com/v1/9c92c43ed76d0d0589914fab726015e08d2defb7",
    "https://rpc-mainnet.maticvigil.com/v1/a2e04be7c65204ab96046fb9a84e36b112ef082e",
    "https://rpc-mainnet.maticvigil.com/v1/287560478b0282eee31a95f3e7da08efdf7571ef",
    "https://rpc-mainnet.maticvigil.com/v1/dcca6097dd8fecd19ee7c07acf8e38b7f5c12809",

]

provider_uris_archival = [
    "https://rpc-mainnet.maticvigil.com/v1/927c67086ea722f1b98b3ec14b174ce140d3a84e",
    "https://rpc-mainnet.maticvigil.com/v1/5dc314d2597052c4d1e00370733d762e5fcd3334",
    "https://rpc-mainnet.maticvigil.com/v1/908b40606eada9119d934ae2809b566b87f61cd7",
    "https://rpc-mainnet.maticvigil.com/v1/058a788a199c0017ce46b3900ddd801c6412372b",
    "https://rpc-mainnet.maticvigil.com/v1/3c91e94ed992e48ae091ed12d1d821f2fbdc0b43",
    "https://rpc-mainnet.maticvigil.com/v1/f9c8eaeeb3c203c8fab2a978c41300b68421dccc",
    "https://rpc-mainnet.maticvigil.com/v1/9c92c43ed76d0d0589914fab726015e08d2defb7",
    "https://rpc-mainnet.maticvigil.com/v1/a2e04be7c65204ab96046fb9a84e36b112ef082e",
    "https://rpc-mainnet.maticvigil.com/v1/287560478b0282eee31a95f3e7da08efdf7571ef",
    "https://rpc-mainnet.maticvigil.com/v1/dcca6097dd8fecd19ee7c07acf8e38b7f5c12809",
]

export_max_workers = min(10,len(provider_uris) - 1)
export_batch_size = 10
output_bucket = "nftbank-matic-network"

# %%
CREDENTIALS = f".cred/pipeline-executor.contxtsio-nftbank.credentials"
os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = CREDENTIALS
SCOPE = ('https://www.googleapis.com/auth/cloud-platform',
        # 'https://www.googleapis.com/auth/bigquery',
        # 'https://www.googleapis.com/auth/drive'
        )
credentials, _ = google.auth.default(scopes=SCOPE)

cloud_storage_hook = SimpleGoogleCloudStorageHook(credentials=credentials, project_id="contxtsio-nftbank")

# %%
task_manager = TaskManager(mongodb_connection_uri="mongodb+srv://masterContxtsio:Contxtsio!23@cluster0.7qjx5.mongodb.net/admin?retryWrites=true&w=majority",
                           min_timestamp=datetime(2020, 5, 30, 16, 0, 0, 0, tzinfo=timezone.utc))
task_manager.initialize_db(TaskIDs['ALL'], "raw")

# %%
# Export
def export_path(directory, start_timestamp):
    return "export/raw/{directory}/block_date={block_date}/block_hour={block_hour}/".format(
        directory=directory,
        block_date=start_timestamp.strftime("%Y-%m-%d"),
        block_hour=start_timestamp.strftime("%H"),
    )

def copy_to_export_path(file_path, export_path):
    logging.info('Calling copy_to_export_path({}, {})'.format(file_path, export_path))
    filename = os.path.basename(file_path)

    cloud_storage_hook.upload_to_gcs(
        gcs_hook=cloud_storage_hook,
        bucket=output_bucket,
        object=export_path + filename,
        filename=file_path)

def copy_from_export_path(export_path, file_path):
    logging.info('Calling copy_from_export_path({}, {})'.format(export_path, file_path))
    filename = os.path.basename(file_path)

    cloud_storage_hook.download_from_gcs(gcs_hook=cloud_storage_hook, bucket=output_bucket, object=export_path + filename, filename=file_path)


# %%
storage_client = cloud_storage_hook.get_conn()

# %%
def add_provider_uri_fallback_loop(python_callable, provider_uris):
    """Tries each provider uri in provider_uris until the command succeeds"""

    def python_callable_with_fallback(**kwargs):
        for index, provider_uri in enumerate(provider_uris):
            kwargs['provider_uri'] = provider_uri
            try:
                ret = python_callable(**kwargs)
                break
            except Exception as e:
                if index < (len(provider_uris) - 1):
                    logging.exception('An exception occurred. Trying another uri')
                else:
                    raise e

        return ret

    return python_callable_with_fallback

# %%
##########################################
### EXTRANCTIONS
##########################################

# %%
def _get_block_range(provider_uri, start_timestamp, end_timestamp, **kwargs):
    with TemporaryDirectory() as tempdir:
        logging.info('Calling get_block_range_for_timestamps ({}, {})'.format(start_timestamp, end_timestamp))
        get_block_range_for_timestamps.callback(
            provider_uri=provider_uri,
            start_timestamp=start_timestamp,
            end_timestamp=end_timestamp,
            output=os.path.join(tempdir, "blocks_meta.txt")
        )

        with open(os.path.join(tempdir, "blocks_meta.txt")) as block_range_file:
            block_range = block_range_file.read()
            start_block, end_block = block_range.split(",")

    return int(start_block), int(end_block)


# %%
def _export_blocks_and_transactions_command(start_timestamp, start_block, end_block, provider_uri, **kwargs):
    with TemporaryDirectory() as tempdir:
        logging.info('Calling export_blocks_and_transactions ({}, {})'.format(start_block, end_block))

        export_blocks_and_transactions.callback(
            start_block=start_block,
            end_block=end_block,
            batch_size=export_batch_size,
            provider_uri=provider_uri,
            rate_limit=200,
            max_workers=export_max_workers,
            blocks_output=os.path.join(tempdir, "blocks.json"),
            transactions_output=os.path.join(tempdir, "transactions.json"),
            revolving=True,
        )

        copy_to_export_path(
                os.path.join(tempdir, "blocks.json"), export_path("blocks", start_timestamp)
        )

        copy_to_export_path(
            os.path.join(tempdir, "transactions.json"), export_path("transactions", start_timestamp)
        )


# %%
def _export_receipts_and_logs_command(start_timestamp, provider_uri, **kwargs):
    with TemporaryDirectory() as tempdir:
        copy_from_export_path(
            export_path("transactions", start_timestamp), os.path.join(tempdir, "transactions.json")
        )

        logging.info('Calling extract_field')
        extract_field.callback(
            input=os.path.join(tempdir, "transactions.json"),
            output=os.path.join(tempdir, "transaction_hashes.txt"),
            field="hash",
        )

        logging.info('Calling export_receipts_and_logs')
        export_receipts_and_logs.callback(
            batch_size=export_batch_size,
            transaction_hashes=os.path.join(tempdir, "transaction_hashes.txt"),
            provider_uri=provider_uri,
            rate_limit=200,
            max_workers=export_max_workers,
            receipts_output=os.path.join(tempdir, "receipts.json"),
            logs_output=os.path.join(tempdir, "logs.json"),
            revolving=True,
        )

        copy_to_export_path(
            os.path.join(tempdir, "receipts.json"), export_path("receipts", start_timestamp)
        )
        copy_to_export_path(
            os.path.join(tempdir, "logs.json"), export_path("logs", start_timestamp)
        )


# %%
def extract_token_transfers_command(execution_date, **kwargs):
    with TemporaryDirectory() as tempdir:
        copy_from_export_path(
            export_path("logs", execution_date), os.path.join(tempdir, "logs.json")
        )

        logging.info('Calling extract_token_transfers(..., {}, ..., {})'.format(
            export_batch_size, export_max_workers
        ))
        extract_token_transfers.callback(
            logs=os.path.join(tempdir, "logs.json"),
            batch_size=export_batch_size,
            output=os.path.join(tempdir, "token_transfers.json"),
            max_workers=export_max_workers,
        )

        copy_to_export_path(
            os.path.join(tempdir, "token_transfers.json"),
            export_path("token_transfers", execution_date),
        )

# %%
##########################################
### Get target ranges
##########################################
for task in task_manager.get_todos(TaskIDs["ALL"], "raw"):
    start_timestamp = task['start_timestamp']
    end_timestamp = task['end_timestamp']
    print(start_timestamp, end_timestamp)

    task_manager.task_started(TaskIDs["ALL"], "raw", start_timestamp, end_timestamp)

    ##########################################
    ### Get start / end block number
    ##########################################
    get_block_range = add_provider_uri_fallback_loop(_get_block_range, provider_uris)
    start_block, end_block = get_block_range(**{
        "start_timestamp": start_timestamp.timestamp(),
        "end_timestamp": end_timestamp.timestamp()
    })

    ##########################################
    ### Export blocks and transactions
    ##########################################
    _export_blocks_and_transactions_command(start_timestamp=start_timestamp,
                                            start_block=start_block,
                                            end_block=end_block,
                                            provider_uri=provider_uris)


    ##########################################
    ### Export receipts and logs
    ##########################################
    _export_receipts_and_logs_command(start_timestamp=start_timestamp,
                                    provider_uri=provider_uris)

    task_manager.task_finished(TaskIDs["ALL"], "raw", start_timestamp, end_timestamp)
